# /usr/bin/env python
# -*- encoding: utf-8 -*-

"""
Various MAB algorithms.
"""

import math
import random as rndm
import numpy as np
from utils import *

class MAB_Algorithm(object):
    """Base class for MAB algorithms"""
    def __init__(self, arms, history, name='General MAB Algorithm'):
        """constructor
        Args:
            arms: list of Arms
            history: 2D list
        """
        self.name = name
        self.history = History(history)
        self.num_arms = len(self.history.history)
        self.arms = arms

    def update_status(self, diff_history):
        """Just a stub method. Use update_status() of each algorithm."""
        pass

    def bulk_update_status(self, history):
        """Just a stub method. Use bulk_update_status() of each algorithm."""
        pass

    def next_choice(self, dry=False):
        """Just a stub method. Use next_choice() of each algorithm."""
        pass

    def pick(self, dry=False):
        """Play the picked arm."""
        pick = self.next_choice()
        reward = self.arms[pick].pick()
        self.history.update(pick, reward)
        return pick, reward

class Oracle(MAB_Algorithm):
    """Always picks the best arm for each round."""
    def __init__(self, arms, history, results, name='Oracle'):
        MAB_Algorithm.__init__(self, arms, history, name=name)
        self.results = History(results)
        self.pointers = [0 for _ in range(0, self.num_arms)]

    def next_choice(self):
        """Pick the best arm for the round."""
        candidates = {str(arm): self.results.history[arm][self.pointers[arm]] for arm, _ in enumerate(self.pointers)}
        pick = max(candidates.iteritems(), key=operator.itemgetter(1))[0]
        return int(pick)

    def pick(self, dry=False):
        pick = self.next_choice()
        reward = self.results.history[pick][self.pointers[pick]]
        self.pointers = [pointer+1 for pointer in self.pointers]
        self.history.update(pick, reward)
        return pick, reward

class Random(MAB_Algorithm):
    """Picks a ramdom arm."""
    def __init__(self, arms, history, name='Random'):
        MAB_Algorithm.__init__(self, arms, history, name=name)

    def next_choice(self):
        """Picks a ramdom arm."""
        pick = np.random.choice(self.num_arms, 1)[0]
        return pick

class Epsilon_First(MAB_Algorithm):
    """Epsilon-First algorithm."""
    def __init__(self, arms, history, plays, epsilon=0.05, name='Epsilon First'):
        """
        Args:
            plays: number of total plays
            epsilon: epsilon
        """
        MAB_Algorithm.__init__(self, arms, history, name=name)
        self.epsilon = epsilon
        self.plays = plays
        self.explorations = int(float(plays)*float(epsilon))
        self.plays_counter = 0

    def next_choice(self):
        """
        Randomly pick an arm for exploration.
        Pick the best arm for exploitation.
        """
        if self.plays_counter < self.explorations:
            pick = np.random.choice(self.num_arms, 1)[0]
        else:
            pick = self.history.get_best_arm()

        self.plays_counter += 1

        return pick

class Epsilon_Greedy(MAB_Algorithm):
    """Epsilon-Greedy algorithm."""
    def __init__(self, arms, history, epsilon=0.12, name='Epsilon Greedy'):
        """
        Args:
            epsilon: epsilon
        """
        MAB_Algorithm.__init__(self, arms, history, name=name)
        self.epsilon = epsilon

    def next_choice(self):
        """
        Pick the best for probability epsilon.
        Randomly pick from rest of arms for probability 1-epsilon
        """
        probabilities = [None for _ in self.arms]
        best = self.history.get_best_arm()
        others = [None for _ in self.arms]
        for i in range(0, len(self.arms)):
            if i == best:
                probabilities[i] = 1.0 - self.epsilon + self.epsilon/float(len(self.arms))
            else:
                probabilities[i] = self.epsilon/float(len(self.arms))

        return pick_by_probability(probabilities)

class Softmax(MAB_Algorithm):
    """Boltzmann Exploration (Softmax) algorithm."""
    def __init__(self, arms, history, tau=0.05, name='Softmax'):
        """
        Args:
            tau: tau
        """
        MAB_Algorithm.__init__(self, arms, history, name=name)
        self.tau = tau

    def next_choice(self):
        unknown_arm = self.history.get_unknown_arm()
        if unknown_arm is None:
            estimated_means = self.history.get_estimated_means()
            divider = sum([math.exp(estimated_mean/self.tau) for estimated_mean in estimated_means])
            probabilities = [math.exp(estimated_mean/self.tau)/divider for estimated_mean in estimated_means]
            pick = pick_by_probability(probabilities)
        else:
            pick = unknown_arm
        return pick

class UCB(MAB_Algorithm):
    """Upper Confidence Bound algorithm."""
    def __init__(self, arms, history, name='UCB'):
        MAB_Algorithm.__init__(self, arms, history, name=name)

    def next_choice(self):
        unknown_arm = self.history.get_unknown_arm()
        if unknown_arm is None:
            estimated_means = self.history.get_estimated_means()
            plays = self.history.get_plays()
            rounds_count = self.history.get_rounds_so_far()

            evaluations = [estimated_means[arm] + math.sqrt(2.0*math.log(float(rounds_count))/float(plays[arm])) for arm in range(0, len(self.arms))]
            pick = evaluations.index(max(evaluations))
        else:
            pick = unknown_arm
        return pick

class Single_Arm(MAB_Algorithm):
    """Keep choosing single arm."""
    def __init__(self, arms, history, designated_arm, name='Single Arm'):
        """
        Args:
            designated_arm: arm to pick everytime
        """
        MAB_Algorithm.__init__(self, arms, history, name=name)
        self.designated_arm = designated_arm
        if self.arms[self.designated_arm].label:
            self.name = name + '(%s)' % self.arms[self.designated_arm].label
        else:
            self.name = name

    def next_choice(self):
        return self.designated_arm

def pick_by_probability(probabilities):
    r = rndm.random()
    pick = None
    for idx, p in enumerate(probabilities):
        r -= p
        if r <= 0.0:
            pick = idx
            break
    return pick



# /usr/bin/env python
# -*- encoding: utf-8 -*-

"""
Multi-Armed Bandits agent.
"""

from algorithms import *
from utils import *

class MAB(object):
    """
    Agent class representing MAB player.
    """
    def __init__(self):
        self.arms = []
        self.history = History([])

    def add_auto_generated_arms(self, num_arms, labels=None):
        self.arms += [Arm(label="Arm-%s" % (len(self.arms)+i)) for i in range(0, num_arms)]
        for _ in range(0, num_arms):
            self.history.add_arm([])

    def add_arm(self, dist_type='nd', label=None, mean=None, sd=None, truncate=[0.0, 1.0]):
        self.arms.append(Arm(dist_type=dist_type, label=label, mean=mean, sd=sd, truncate=truncate))
        self.history.add_arm([])

    def set_algorithm(self, algorithm):
        self.algorithm = algorithm

    def play(self):
        pick, reward = self.algorithm.pick()

    def summary(self):
        summary = {}
        summary['algorithm'] = self.algorithm.name
        summary['rounds'] = self.history.get_rounds_so_far()
        summary['true_means'] = [arm.mean for arm in self.algorithm.arms]
        summary['true_sds'] = [arm.sd for arm in self.algorithm.arms]
        summary['empirical_means'] = self.history.get_means()
        summary['empirical_sds'] = self.history.get_sds()
        summary['regret'] = self.calc_regret()
        summary['total_reward'] = self.history.get_total_reward()
        summary['plays'] = [len(hist_of_arm) for hist_of_arm in self.history.get_dense_history()]
        return summary

    def calc_regret(self):
        # Compare to Oracle
        oracle = MAB()
        results = [[arm.pick() for _ in range(0, self.history.get_rounds_so_far())] for arm in self.arms]
        for arm in self.arms:
            oracle.add_arm(label=arm.label, mean=arm.mean, sd=arm.sd, truncate=arm.truncate)
        oracle.set_algorithm(Oracle(arms=self.arms, history=oracle.history.history, results=results))
        for _ in range(0, self.history.get_rounds_so_far()):
            oracle.play()

        oracle_total_reward = oracle.history.get_total_reward()
        total_reward = self.history.get_total_reward()
        regret = oracle_total_reward - total_reward
        return regret

class Simulated_MAB(MAB):
    def __init__(self, rounds):
        MAB.__init__(self)
        self.rounds = rounds

    def add_auto_generated_arms(self, num_arms, labels=None):
        self.arms += [Simulated_Arm(self.rounds, label="Arm-%s" % (len(self.arms)+i)) for i in range(0, num_arms)]
        for _ in range(0, num_arms):
            self.history.add_arm([])

    def add_arm(self, dist_type='nd', label=None, mean=None, sd=None, truncate=[0.0, 1.0]):
        self.arms.append(Simulated_Arm(self.rounds, dist_type=dist_type, label=label, mean=mean, sd=sd, truncate=truncate))
        self.history.add_arm([])

    def reset(self, dry=True):
        self.history.reset()
        for arm in self.arms:
            arm.reset(dry=dry)

    def calc_regret(self):
        # Compare to Oracle
        oracle = Simulated_MAB(self.history.get_rounds_so_far())
        # results = [[arm.pick() for _ in range(0, self.history.get_rounds_so_far())] for arm in self.arms]
        results = [arm.results for arm in self.arms]
        for arm in self.arms:
            oracle.add_arm(label=arm.label, mean=arm.mean, sd=arm.sd, truncate=arm.truncate)
        for idx, arm in enumerate(oracle.arms):
            arm.force_update_results([results[idx]])
        oracle.set_algorithm(Oracle(arms=self.arms, history=oracle.history.history, results=results))
        for _ in range(0, self.history.get_rounds_so_far()):
            oracle.play()

        oracle_total_reward = oracle.history.get_total_reward()
        total_reward = self.history.get_total_reward()
        regret = oracle_total_reward - total_reward
        return regret

class InvalidSettingException(Exception):
    """
    Exception for invalid settings of Arm
    """
    def __init(self, msg='Failed to update history.'):
        Exception.__init__(self, msg)



# /usr/bin/env python
# -*- encoding: utf-8 -*-

# Simulation of Multi-armed Bandits

import sys, math, json
from matplotlib import pyplot as plt
import numpy as np
from scipy.stats import truncnorm
import random as rndm
import collections
# matplotlib.use('TkAgg')
# plt.rcParams["figure.figsize"] = (6, 4)


# ## Class: Arm

# In[188]:

# Constants
INF = 10**5
PRESET_RESULTS = INF
FILE_OUTPUT = False


# In[189]:

class Arm():
    def __init__(self, distribution='nd', label=None, mean=None, sd=None):
        self.dist = distribution
        self.mean = self.sigma = None
        while self.mean is None or self.mean > 1.0 or self.mean < .0:
            self.mean = mean or np.random.normal(.5, .1, 1)[0]
        while self.sigma is None or self.sigma > 1.0 or self.sigma < .0:
#             self.sigma = sd or np.random.normal(.1, .1, 1)[0]
            self.sigma = 0.25
        self.label = label
        self.results = [ e for e in np.random.normal(self.mean, self.sigma, PRESET_RESULTS) if e >= 0.0 and e <= 1.0 ]
#         self.results = [ v+self.mean for v in truncnorm.rvs(-self.mean, 1.0-self.mean, scale=1.0, size=PRESET_RESULTS) ]
        self.pointer = 0

    def pick(self):
        if self.pointer >= PRESET_RESULTS: raise ValueError("No more results.")
        reward = self.results[self.pointer]
        while reward < 0.0 or reward > 1.0:
            self.pointer += 1
            reward = self.results[self.pointer]
#         if reward > 1.0: reward = 1.0
#         if reward < 0.0: reward = 0.0
#         self.pointer += 1
        return reward



# ## Class: MAB

# In[190]:

class MAB():
#     def __init__(self, arms, distribution='nd'):
#         self.k = arms
#         self.dist = distribution
#         self.arms = [Arm(distribution=self.dist) for _ in range(0, self.k)]
#         if FILE_OUTPUT:
#             with open("./preset_results.csv", 'a') as f:
#                 for j in range(0, 500):
#                     values = [str(self.arms[i].results[j]) for i in range(0, len(self.arms))]
#                     record = ",".join(values)
#                     f.write(record + "\n")

    def __init__(self, arms, distribution='nd'):
        self.arms = arms
        self.dist = distribution
        self.arms = [Arm(distribution=self.dist, label=a['label'] or None, mean=a['mean'] or None, sd=a['sd'] or None) for a in self.arms]
        if FILE_OUTPUT:
            with open("./preset_results.csv", 'a') as f:
                for j in range(0, 500):
                    values = [str(self.arms[i].results[j]) for i in range(0, len(self.arms))]
                    record = ",".join(values)
                    f.write(record + "\n")

    def pick(self, i):
        # Pick ith arm
        reward = self.arms[i].pick()
        for arm in self.arms:
            arm.pointer += 1
        return reward

    def draw_arms_histogram(self, i=None):
        if i:
            x = self.arms[i].results
            plt.hist(x)
            plt.figure()
        else:
            for arm in self.arms:
                x = arm.results
                plt.hist(x)
                plt.figure()

    def combinations(self, limit=None):
        # Compute number of combinations with less than L arms among K arms
        # arguments
        # * limit: Limitation of the number of arms that consist a super-arm
        # * draw: Option to draw graph
        # * table: Option to show table

        if limit and type(limit)!=int: return "L should be int or None."

        combs = {}
        f = math.factorial
        combs = {i:f(self.k)/f(i)/f(self.k-i) for i in range(1, self.k+1)}

        if limit and type(limit)==int:
            if limit > self.k: return "L should be smaller than or equal to K."
            if limit < 0: return 0
            combsum = 0
            for k,v in combs.iteritems():
                if k <= limit: combsum += v
            return combsum
        else:
            return combs

    def draw_combinations(self):
        # Draw number of combinations for each possible number of arms
        x = np.linspace(1, self.k, self.k, endpoint=True)
        y = [self.combinations(i) for i in range(1, self.k+1)]
        plt.xlabel('Number of incentives')
        plt.ylabel('Nuber of combinations')
        plt.plot(x,y)
#         plt.figure()
#         plt.draw()
        plt.savefig('combinations.png', format='png', dpi=960)
        plt.show()
#         plt.close()
        return None


# ## Class: MAB_Algorithm

# In[191]:

class MAB_Algorithm(object):
    def __init__(self, mab):
        self.name = 'MAB Algorithm'
        self.mab = mab
        for arm in self.mab.arms:
            arm.pointer = 0
        ranking = [8, 1, 3, 2, 5, 6, 7, 9, 4] # 実際の利得の順位
        # 期待利得の初期値を0で始める場合
        self.status = { i:{ "estimated_mean": .0, "total_reward": .0, "stddev": .0, "results": [] } for i in range(0, len(self.mab.arms)) }
        # 期待利得の初期値を1で始める場合
#         self.status = { i:{ "estimated_mean": 1.0, "total_reward": .0, "stddev": .0, "results": [] } for i in range(0, len(self.mab.arms)) }
        # 期待利得分布の初期値に，実際の利得分布と同じ値を与える場合
#         self.status = { i:{ "estimated_mean": self.mab.arms[i].mean, "total_reward": .0, "stddev": .0, "results": [] } for i in range(0, len(self.mab.arms)) }
        # 期待利得分布の初期値に，実際の利得分布と異なる値で平均値の順序のみを保存したランダム値を与える場合
#         self.status = { i:{ "estimated_mean": np.random.normal(0.75 - 0.05*ranking[i], .1, 1)[0], "total_reward": .0, "stddev": .0, "results": [] } for i in range(0, len(self.mab.arms)) }
        # 期待利得分布の初期値に，実際の利得分布と異なる値で平均値の順序のみを保存したランダム値(平均値の分散が小さい)を与える場合
#         self.status = { i:{ "estimated_mean": np.random.normal(0.6 - 0.01*ranking[i], .1, 1)[0], "total_reward": .0, "stddev": .0, "results": [] } for i in range(0, len(self.mab.arms)) }
        # 期待利得分布の初期値に，実際の利得分布と異なる値で平均値の上位半分，下位半分のみ保存したランダム値を与える場合
        # 期待利得分布の初期値に，ランダム値を与える場合
#         self.status = { i:{ "estimated_mean": np.random.normal(0.5, .1, 1)[0], "total_reward": .0, "stddev": .0, "results": [] } for i in range(0, len(self.mab.arms)) }

        self.played_rounds = 0

    def make_decision(self):
        return None

    def play(self):
        arm = self.make_decision()
        reward = self.mab.pick(arm)
        if FILE_OUTPUT:
            with open("./%s.csv" % self.name, 'a') as f:
                f.write("%s,%s,%s\n" % (self.mab.arms[arm].pointer, arm, reward))
        return { "arm": arm, "reward": reward }

    def simulate(self, rounds):
        self.rounds = rounds
        for r in range(0, rounds):
            self.played_rounds = r
            # Play this round
            result = self.play()
            arm = result["arm"]
            reward = result["reward"]

            # Update results
            self.status[arm]["results"].append({r: reward})

            # Update other statuses
            total = .0
            for v in self.status[arm]["results"]:
                total += float(v.values()[0])
            self.status[arm]["total_reward"] = float(total)
            self.status[arm]["estimated_mean"] = float(total)/float(len(self.status[arm]["results"]))
            var = .0
            for v in self.status[arm]["results"]:
                var += (self.status[arm]["estimated_mean"] - float(v.values()[0]))**2
            self.status[arm]["stddev"] = math.sqrt(var/float(len(self.status[arm]["results"])))

    def result(self):
        result = {}
        result["total_reward"] = sum([v["total_reward"] for k, v in self.status.iteritems()])
        result["status"] = self.status
        return result

    def show_result(self):
        print "Arm\tTotal Reward\tEstimated Mean\tTrue Mean\tPlayed Rounds"
        for k, v in self.status.iteritems():
            print "%s\t%s\t%s\t%s\t%s" % (k, v["total_reward"], v["estimated_mean"], self.mab.arms[k].mean, len(v["results"]))
        print "Total Reward = %s" % self.result()["total_reward"]

    def graph_result(self):
        estimated_means = [v["estimated_mean"] for k, v in self.status.iteritems()]
        std_estimated_mean = [v["stddev"] for k, v in self.status.iteritems()]
        true_means = [m.mean for m in self.mab.arms]

        fig, ax = plt.subplots()
        index = np.arange(len(self.mab.arms))
        bar_width = 0.4
        alpha = 0.5

        rects1 = plt.bar(index, estimated_means, bar_width, color='b', alpha=alpha, yerr=std_estimated_mean, label="Estimated Mean")
        rects2 = plt.bar(index + bar_width, true_means, bar_width, color='r', alpha=alpha, label="True Mean")

        plt.xlabel('Arm')
        plt.ylabel('Reward')
        plt.title('Mean reward of each arm')
        plt.xticks(index + bar_width, [str(k) for k, v in self.status.iteritems()])
        plt.legend()

#         plt.tight_layout()
        plt.show()


# In[192]:

# Oracle(Always picks the best arm at each time)
class Oracle(MAB_Algorithm):
    def __init__(self, mab):
        super(Oracle, self).__init__(mab)
        self.name = 'oracle'
    def true_best_arm(self):
        max_mean = -INF
        selection = None
        for k, v in self.status.iteritems():
            arm = self.mab.arms[k]
            if arm.results[arm.pointer] > max_mean:
                selection = k
                max_mean = arm.results[arm.pointer]
        return selection

    def make_decision(self):
        return self.true_best_arm()



# In[193]:

# Random Algorithm
class Random(MAB_Algorithm):
    def __init__(self, mab):
        super(Random, self).__init__(mab)
        self.name = 'random'
    def make_decision(self):
        return np.random.choice(len(self.mab.arms), 1)[0]


# In[194]:

# Epsilon First Algorithm
class Epsilon_First(MAB_Algorithm):
    def __init__(self, mab, epsilon=.5):
        super(Epsilon_First, self).__init__(mab)
        self.name = 'epsilon_first'
        self.epsilon = epsilon

    def best_arm(self):
        max_mean = -INF
        selection = None
        for k,v in self.status.iteritems():
            if v["estimated_mean"] > max_mean:
                selection = k
                max_mean = v["estimated_mean"]
        return selection

    def make_decision(self):
        if float(self.played_rounds) < float(self.rounds)*self.epsilon:
            return np.random.choice(len(self.mab.arms), 1)[0]
        else:
            return self.best_arm()


# In[195]:

# Epsilon-Greedy Algorithm
class Epsilon_Greedy(MAB_Algorithm):
    def __init__(self, mab, epsilon=.12):
        super(Epsilon_Greedy, self).__init__(mab)
        self.name = 'epsilon_greedy'
        self.epsilon = epsilon

#     def best_arm(self):
#         max_eval = float(-INF)
#         for k, v in self.status.iteritems():
#             evaluation = v["estimated_mean"]
#             if evaluation > max_eval:
#                 max_eval = evaluation
#                 selection = k
#         return k

    def best_arm(self):
        max_mean = -INF
        selection = None
        for k,v in self.status.iteritems():
            if v["estimated_mean"] > max_mean:
                selection = k
                max_mean = v["estimated_mean"]
        return selection

    def make_decision(self):
#         for i in range(0, len(self.mab.arms)):
#             if len(self.status[i]["results"]) == 0:
#                 return i
        best = self.best_arm()
        others = []
        for k, v in self.status.iteritems():
            if k==best:
                v["probability"] = 1.0 - self.epsilon + self.epsilon/float(len(self.mab.arms))
            else:
                others.append(k)
                v["probability"] = self.epsilon/float(len(self.mab.arms))

        r = rndm.random()
        if r <= self.status[best]['probability']:
            selection = best
        else:
            selection = others[int((r - self.status[best]['probability'])/(self.epsilon/float(len(self.mab.arms))))]
        return selection


# In[196]:

# Boltzmann Exploration (Softmax)
class Softmax(MAB_Algorithm):
    def __init__(self, mab, tau=.5):
        super(Softmax, self).__init__(mab)
        self.name = 'softmax'
        self.tau = float(tau)

    def make_decision(self):
        for i in range(0, len(self.mab.arms)):
            if len(self.status[i]["results"]) == 0:
                return i

        divider = sum([ math.exp(v["estimated_mean"]/self.tau) for k, v in self.status.iteritems() ])

        for k, v in self.status.iteritems():
            v["probability"] = math.exp(v["estimated_mean"]/self.tau)/divider

        r = rndm.random()
        selection = None
        for k, v in self.status.iteritems():
            r -= v["probability"]
            if r <= 0.0:
                selection = k
                break
        return selection


# In[197]:

# Pursuit Algorithms


# In[198]:

# Reinforcement Comparison


# In[199]:

# UCB Algorithm
class UCB(MAB_Algorithm):
    def __init__(self, mab):
        super(UCB, self).__init__(mab)
        self.name = 'ucb'

    def make_decision(self):
        for i in range(0, len(self.mab.arms)):
            if len(self.status[i]["results"]) == 0:
                return i
        max_eval = float(-INF)
        selection = None
        for k, v in self.status.iteritems():
            evaluation = v["estimated_mean"] + math.sqrt(float(2.0*math.log(float(self.played_rounds)))/float(len(v["results"])))
            if evaluation > max_eval:
                max_eval = evaluation
                selection = k
        return selection


# In[200]:

# Single Incentive
class Single_Incentive(MAB_Algorithm):
    def __init__(self, mab, label):
        super(Single_Incentive, self).__init__(mab)
        self.name = label
    def make_decision(self):
        for idx, arm in enumerate(self.mab.arms):
            if arm.label == self.name:
                return idx


# ## Simulation
#
# Algorithms: From "Algorithms for the multi-armed bandit problem"

# ### Initialization

# In[201]:

# K = 13 # アーム数
T = 1000 # ラウンド数
# mab = MAB(K)

# 予備実験に基づくシミュレーション
incentives = []

# Q3
incentives.append({ "label": "Bayesian Truth Serum", "mean": 0.300, "sd": None })
incentives.append({ "label": "Punishment Accuracy", "mean": 0.737, "sd": None })
incentives.append({ "label": "Reward Accuracy", "mean": 0.487, "sd": None })
incentives.append({ "label": "Trust", "mean": 0.507, "sd": None })
incentives.append({ "label": "Humanization", "mean": 0.414, "sd": None })
incentives.append({ "label": "Solidarity", "mean": 0.400, "sd": None })
incentives.append({ "label": "Cheap Talk - Normative", "mean": 0.350, "sd": None })
incentives.append({ "label": "Cheap Talk - Survaillance", "mean": 0.233, "sd": None })
incentives.append({ "label": "Tournament Scoring", "mean": 0.425, "sd": None })

# Q4
# incentives.append({ "label": "Bayesian Truth Serum", "mean": 0.520, "sd": None })
# incentives.append({ "label": "Punishment Accuracy", "mean": 0.743, "sd": None })
# incentives.append({ "label": "Reward Accuracy", "mean": 0.487, "sd": None })
# incentives.append({ "label": "Trust", "mean": 0.750, "sd": None })
# incentives.append({ "label": "Humanization", "mean": 0.575, "sd": None })
# incentives.append({ "label": "Solidarity", "mean": 0.555, "sd": None })
# incentives.append({ "label": "Cheap Talk - Normative", "mean": 0.500, "sd": None })
# incentives.append({ "label": "Cheap Talk - Survaillance", "mean": 0.600, "sd": None })
# incentives.append({ "label": "Tournament Scoring", "mean": 0.520, "sd": None })

trials = [ MAB(incentives) for _ in range(0, 10) ]
results_simulation = [ {} for _ in range(0, 10) ]


# In[202]:

# 初期状態の妥当性チェック
# mab.draw_arms_histogram()


# In[203]:

# Oracle
for idx, mab in enumerate(trials):
    oracle = Oracle(mab)
    oracle.simulate(T)
    oracle.show_result()
    oracle.graph_result()
    results_simulation[idx]['Oracle'] = oracle.result()['total_reward']


# In[204]:

# Random
for idx, mab in enumerate(trials):
    random = Random(mab)
    random.simulate(T)
    random.show_result()
    random.graph_result()
    results_simulation[idx]['Random'] = random.result()['total_reward']
# results_simulation[''] = []
# for mab in trials:
#     results_simulation[''].append(oracle.result()['total_reward'])


# In[205]:

# Epsilon First

label_ef = []
# result_ef = []
for idx, mab in enumerate(trials):
    EPSILON = .04
    for i in range(1, 14):
        label = "Epsilon First(e=%s)" % (EPSILON*float(i)*20.0)
        if label not in label_ef: label_ef.append(label)
        epsilon_first = Epsilon_First(mab, EPSILON*float(i)*20.0)
        epsilon_first.simulate(T)
        epsilon_first.show_result()
        epsilon_first.graph_result()
        results_simulation[idx][label] = epsilon_first.result()['total_reward']

#     label_ef.append("Epsilon First(e=%s)" % (EPSILON*float(i)))
#     result_ef.append(epsilon_first.result()["total_reward"])

# results_simulation[''] = []
# for mab in trials:
#     results_simulation[''].append(oracle.result()['total_reward'])


# In[206]:

# Epsilon Greedy
# EPSILON = 0.5
# epsilon_greedy = Epsilon_Greedy(mab, EPSILON)
# epsilon_greedy.simulate(T)
# epsilon_greedy.show_result()
# epsilon_greedy.graph_result()


label_eg = []
result_eg = []
for idx, mab in enumerate(trials):
    EPSILON = .06
    for i in range(1, 14):
        label = "Epsilon Greedy(e=%s)" % (EPSILON*float(i)*20.0)
        if label not in label_eg: label_eg.append(label)
        epsilon_greedy = Epsilon_Greedy(mab, EPSILON*float(i)*20.0)
        epsilon_greedy.simulate(T)
        epsilon_greedy.show_result()
        epsilon_greedy.graph_result()
        results_simulation[idx][label] = epsilon_greedy.result()['total_reward']
#         label_eg.append("Epsilon Greedy(e=%s)" % (EPSILON*float(i)))
#         result_eg.append(epsilon_greedy.result()["total_reward"])


# In[207]:

# Boltzmann Exploration (Softmax)
label_sm = []
for idx, mab in enumerate(trials):
    TAU = .04
    for i in range(1, 14):
        label = "Softmax(t=%s)" % (TAU*float(i)*20.0)
        if label not in label_sm: label_sm.append(label)
        softmax = Softmax(mab, TAU*float(i)*20.0)
        softmax.simulate(T)
        softmax.show_result()
        softmax.graph_result()
        results_simulation[idx][label] = softmax.result()['total_reward']
#         label_sm.append("Softmax(t=%s)" % (TAU*float(i)))
#         result_sm.append(softmax.result()["total_reward"])


# In[208]:

# UCB
for idx, mab in enumerate(trials):
    ucb = UCB(mab)
    ucb.simulate(T)
    ucb.show_result()
    ucb.graph_result()
    results_simulation[idx]["UCB"] = ucb.result()['total_reward']


# In[209]:

# Single Incentives
label_si = []
for idx, mab in enumerate(trials):
    for incentive in incentives:
        label = "Single(%s)" % incentive['label']
        if label not in label_si: label_si.append(label)
        single = Single_Incentive(mab, label=incentive['label'])
        single.simulate(T)
        single.show_result()
        single.graph_result()
        results_simulation[idx][label] = single.result()['total_reward']


# ### 複数回(K=9, T=1000, 100回繰り返しで平均値)

# In[210]:

# print results_simulation
with open('./simulation/output/result.json', 'w') as f:
    f.write(json.dumps(results_simulation, indent=2))
    f.write('\n')

plt.rcParams["figure.figsize"] = (12, 20)
srs = [ collections.OrderedDict(sorted(r.items())) for r in results_simulation ]  # sorted results_simulation

x = ["Oracle", "Random"] + label_si + label_ef + label_eg + label_sm + ["UCB"]
summary = {}
for incentive in x:
    sum_reward = sum([ res[incentive] for res in srs ])
    mean_reward = sum_reward/float(len(srs))
    sum_sq = sum([ (mean_reward-res[incentive])**2.0 for res in srs ])
    sd = math.sqrt(sum_sq)/float(len(srs))
    summary[incentive] = {'mean':mean_reward, 'sd':sd}

with open('./simulation/output/summary.json', 'w') as f:
    f.write(json.dumps(summary, indent=2))
    f.write('\n')

y = [ summary[incentive]['mean'] for incentive in x ]
yerr = [ summary[incentive]['sd'] for incentive in x ]
index = np.arange(len(x))

bar_width = 0.7
alpha = 0.5

fig, ax = plt.subplots()
for i, v in enumerate(y):
    ax.text(v + 10, i + .1, round(v, 1), color='black')

# plt.rcParams["figure.figsize"] = (12, 10)
barlist = plt.barh(index, y, bar_width, xerr=yerr, alpha=alpha, label="Total Reward")
for idx, bar in enumerate(barlist):
    if x[idx].find("Oracle") >= 0: bar.set_color('r')
    elif x[idx].find("Random") >= 0: bar.set_color('r')
    elif x[idx].find("Single") >= 0: bar.set_color('g')
    elif x[idx].find("Epsilon First") >= 0: bar.set_color('b')
    elif x[idx].find("Epsilon Greedy") >= 0: bar.set_color('g')
    elif x[idx].find("Softmax") >= 0: bar.set_color('b')
    elif x[idx].find("UCB") >= 0: bar.set_color('g')
barlist[0].set_color('r')
plt.ylabel('Algorithm')
plt.xlabel('Total Reward')
plt.title('Total reward of each algorithm')
plt.yticks(index+bar_width/2.0, x)

fig_size = plt.rcParams["figure.figsize"]

plt.figure()
plt.show()


# ### グラフが大きすぎるので分割

# In[211]:

# print results_simulation
# with open('./simulation/output/result.json', 'w') as f:
#     f.write(json.dumps(results_simulation, indent=2))
#     f.write('\n')

# with open('./simulation/output/summary.json', 'w') as f:
#     f.write(json.dumps(summary, indent=2))
#     f.write('\n')

# アウトプットしたファイルからJSON読み込み
with open('./simulation/output/summary.json', 'r') as f:
    summary = json.loads(f.read())
# print(summary)

plt.rcParams["figure.figsize"] = (8, 6)
srs = [ collections.OrderedDict(sorted(r.items())) for r in results_simulation ]  # sorted results_simulation

x = ["Oracle", "Random"] + label_si + ["Epsilon First(e=0.04)", "Epsilon Greedy(e=0.06)", "Softmax(t=0.04)", "UCB"]
# xs = []
# xs.append(["Oracle", "Random"] + label_si)
# xs.append(["Oracle", "Random"] + ["Epsilon First(e=0.04)"])
# xs.append(["Oracle", "Random"] + ["Epsilon Greedy(e=0.12)"])
# xs.append(["Oracle", "Random"] + ["Softmax(e=0.04)"])
# xs.append(["Oracle", "Random"] + ["UCB"])

# for x in xs:
y = [ summary[incentive]['mean'] for incentive in x ]
yerr = [ summary[incentive]['sd'] for incentive in x ]
index = np.arange(len(x))

bar_width = 0.7
alpha = 0.5

fig, ax = plt.subplots()
for i, v in enumerate(y):
    ax.text(v + 10, i + .1, round(v, 1), color='black')

# plt.rcParams["figure.figsize"] = (12, 10)
barlist = plt.barh(index, y, bar_width, color='g', xerr=yerr, alpha=alpha, label="Total Reward")
for idx, bar in enumerate(barlist):
    if x[idx].find("Oracle") >= 0: bar.set_color('r')
    elif x[idx].find("Random") >= 0: bar.set_color('r')
    elif x[idx].find("Single") >= 0: bar.set_color('g')
    else: bar.set_color('b')
barlist[0].set_color('r')
plt.ylabel('Algorithm')
plt.xlabel('Total Reward')
plt.title('Total reward of each algorithm')
plt.yticks(index+bar_width/2.0, x)
plt.tight_layout()

fig_size = plt.rcParams["figure.figsize"]

plt.savefig('./simulation/output/simulation_selected.png', format='png', dpi=500)
plt.figure()
# plt.show()


# ### 各アルゴリズムの合計利得

# In[ ]:

# # FIXED: OracleよりEpsilon Firstのほうが良い場合があるので直す
# # x = ["Oracle", "Random", "Epsilon First(e=0.1)", "Epsilon Greedy(e=0.1)", "UCB"]
# x = ["Oracle", "Random"] + label_ef + label_eg + label_sm + ["UCB"]
# index = np.arange(len(x))
# # y = [oracle.result()["total_reward"], random.result()["total_reward"], epsilon_first.result()["total_reward"], epsilon_greedy.result()["total_reward"], ucb.result()["total_reward"]]
# y = [oracle.result()["total_reward"], random.result()["total_reward"]] + result_ef + result_eg + result_sm + [ucb.result()["total_reward"]]

# bar_width = 0.7
# alpha = 0.5

# fig, ax = plt.subplots()
# for i, v in enumerate(y):
#     ax.text(v + 3, i + .1, round(v, 1), color='black')
# #     ax.text(v + 3, i + .25, str(v), color='blue', fontweight='bold')

# plt.rcParams["figure.figsize"] = (12, 30)
# barlist = plt.barh(index, y, bar_width, alpha=alpha, label="Total Reward")
# # barlist = plt.barh(index, y, bar_width, color='b', alpha=alpha, label="Total Reward")
# for idx, bar in enumerate(barlist):
#     if x[idx].find("Oracle") >= 0: bar.set_color('r')
#     elif x[idx].find("Random") >= 0: bar.set_color('r')
#     elif x[idx].find("Epsilon First") >= 0: bar.set_color('b')
#     elif x[idx].find("Epsilon Greedy") >= 0: bar.set_color('g')
#     elif x[idx].find("Softmax") >= 0: bar.set_color('b')
#     elif x[idx].find("UCB") >= 0: bar.set_color('g')
# barlist[0].set_color('r')
# plt.ylabel('Algorithm')
# plt.xlabel('Total Reward')
# plt.title('Total reward of each algorithm')
# plt.yticks(index+bar_width/2.0, x)

# fig_size = plt.rcParams["figure.figsize"]
# print "Current size:", fig_size

# # Set figure width to 12 and height to 9
# # fig_size[0] = 12
# # fig_size[1] = 9

# plt.figure()
# plt.show()


# ### 組み合わせ数
#
# $\Sigma^{K}_{i=1} {}_KC_i$

# In[ ]:

# K = 13
# L = 13

# mab = MAB(K)

# print mab.combinations(limit=L)
# mab.draw_combinations()


# In[ ]:



# /usr/bin/env python
# -*- encoding: utf-8 -*-

"""
Utility classes.
"""

import operator
import math
import numpy as np

DEFAULT_MEAN_OF_SD = 0.05
DEFAULT_SD_OF_SD = 0.01
DEFAULT_MEAN_OF_MEAN = 0.5
DEFAULT_SD_OF_MEAN = 0.1

class History(object):
    """
    Class for choice history and all true results(used for simulation).
    """
    def __init__(self, hist):
        """N x M list of past choices
        N: Number of arms
        M: Number of rounds
        history[n][m] is float value only if arm n was picked at round m.
        """
        is_hist_list = isinstance(hist, list)
        is_hist_of_every_arm_list = all([isinstance(hist_of_arm, list) for hist_of_arm in hist])
        is_all_float_or_none = all([element is None or isinstance(element, float) for arm in hist for element in arm])
        is_size_valid = all([len(hist[i]) == len(hist[i-1]) for i in range(0, len(hist))])
        if is_hist_list and is_hist_of_every_arm_list and is_all_float_or_none and is_size_valid:
            self.history = hist
        else:
            raise InvalidHistoryException()

    def reset(self):
        new_history = [[] for _ in self.history]
        self.history = new_history

    def update(self, index_arm, value):
        """
        Update history of a designated arm with given value.
        """
        if isinstance(value, float):
            for i in range(0, len(self.history)):
                if index_arm == i:
                    self.history[index_arm].append(value)
                else:
                    self.history[i].append(None)
        else:
            raise UpdateFailedException()

    def add_arm(self, hist_of_arm, index=None):
        self.history.append(hist_of_arm)

    def get_unknown_arm(self):
        unknown = None
        for idx, hist_of_arm in enumerate(self.get_dense_history()):
            if len(hist_of_arm) == 0:
                unknown = idx
                break
        return unknown

    def get_best_arm(self):
        estimated_means = self.get_estimated_means()
        candidates = {str(idx): estimated_mean for idx, estimated_mean in enumerate(estimated_means)}
        best = int(max(candidates.iteritems(), key=operator.itemgetter(1))[0])
        return best

    def get_estimated_means(self):
        dense_history = self.get_dense_history()
        estimated_means = [float(sum(hist_of_arm))/float(len(hist_of_arm)) if len(hist_of_arm)>0 else 0.0 for hist_of_arm in dense_history]
        return estimated_means

    def get_dense_history(self):
        """
        Remove None elements from history.
        """
        dense_history = []
        for hist_of_arm in self.history:
            dense_history.append([element for element in hist_of_arm if element is not None])
        return dense_history

    def get_rounds_so_far(self):
        return len(self.history[0])

    def get_plays(self):
        return [len(hist_of_arm) for hist_of_arm in self.get_dense_history()]

    def get_total_rewards(self):
        return [sum(hist_of_arm) for hist_of_arm in self.get_dense_history()]

    def get_total_reward(self):
        return sum(self.get_total_rewards())

    def get_means(self):
        plays = self.get_plays()
        total_rewards = self.get_total_rewards()
        return [float(total_reward)/float(play) if play>0 else 0.0 for total_reward, play in zip(total_rewards, plays)]

    def get_sds(self):
        plays = self.get_plays()
        means = self.get_means()
        variances = [sum([(mean-reward)**2.0 for reward in hist_of_arm])/float(play) if play>0 else 0.0 for mean, hist_of_arm, play in zip(means, self.get_dense_history(), plays)]
        sds = [math.sqrt(variance) for variance in variances]
        return sds

class Arm(object):
    """
    Arm class representing each arm.
    """
    def __init__(self, dist_type='nd', label=None, mean=None, sd=None, truncate=[0.0, 1.0]):
        self.mean = mean
        self.sd = sd
        self.label = label
        self.truncate = truncate
        while self.sd is None:
            self.sd = sd or np.random.normal(DEFAULT_MEAN_OF_SD, DEFAULT_SD_OF_SD, 1)[0]
        while self.mean is None or self.mean > self.truncate[1] or self.mean < self.truncate[0]:
            self.mean = mean or np.random.normal(DEFAULT_MEAN_OF_MEAN, DEFAULT_SD_OF_MEAN, 1)[0]

    def update_label(self, label):
        self.label = label

    def pick(self):
        """
        Return a reward from distribution.
        """
        reward = None
        is_valid_value = (reward >= self.truncate[0] or reward <= self.truncate[1])
        while reward is None or not is_valid_value:
            reward = np.random.normal(self.mean, self.sd, 1)[0]
        return reward

class Simulated_Arm(Arm):
    """
    Simulated_Arm class representing each arm with fixed results.
    """
    def __init__(self, size, dist_type='nd', label=None, mean=None, sd=None, truncate=[0.0, 1.0]):
        Arm.__init__(self, dist_type='nd', label=label, mean=mean, sd=sd, truncate=truncate)
        self.size = size
        self.pointer = 0

        self.results = []
        while len(self.results) < self.size:
            reward = np.random.normal(self.mean, self.sd, 1)[0]
            if reward >= 0.0 and reward <= 1.0:
                self.results.append(reward)

    def reset(self, dry=True):
        self.pointer = 0
        if not dry:
            self.results = []
            while len(self.results) < self.size:
                reward = np.random.normal(self.mean, self.sd, 1)[0]
                if reward >= 0.0 and reward <= 1.0:
                    self.results.append(reward)

    def update_state(self):
        self.pointer += 1

    def force_update_results(self, results):
        self.results = results

    def pick(self):
        """
        Return a reward from results list.
        """
        if self.pointer >= self.size:
            reward = None
        else:
            reward = self.results[self.pointer]
            self.pointer += 1
        return reward

class InvalidHistoryException(Exception):
    """
    Exception for invalid values or type of history.
    """
    def __init__(self, msg='Invalid history.'):
        Exception.__init__(self, msg)

class UpdateFailedException(Exception):
    """
    Exception for invalid use of update()
    """
    def __init(self, msg='Failed to update history.'):
        Exception.__init__(self, msg)



from setuptools import setup, find_packages
version = '0.1.11'

# try:
#     import pypandoc
#     read_md = lambda f: pypandoc.convert(f, 'rst')
# except ImportError:
#     print("warning: pypandoc module not found, could not convert Markdown to RST")
#     read_md = lambda f: open(f, 'r').read()
#
setup(
        name='mab',
        version=version,
        description="Multi-Armed Bandits Algorithms",
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Environment :: Console",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "Natural Language :: English",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: CGI Tools/Libraries",
            "Topic :: Utilities",
            "License :: OSI Approved :: MIT License",
            ],
        keywords='MAB, Multi-Armed Bandits, Machine Learning, Reinforcement Learning',
        author='Akihiko ITOH',
        author_email='itoh.akihiko.5@facebook.com',
        url='https://github.com/AkihikoITOH/MAB',
        license='MIT',
        packages=find_packages(exclude=['examples', 'tests', 'locals']),
        include_package_data=True,
        zip_safe=True,
        # long_description=read_md('README.md'),
        install_requires=['numpy', 'pypandoc'],
    )
